from .task.base import AbstractTask

import logging
logger = logging.getLogger(__name__)


def run_tasks_chain(tasks_chain):
    """
    :type tasks_chain: list[AbstractTask]
    :rtype: str
    """

    for task in tasks_chain:
        logger.info("Executing {task}. output: {output}".format(
            task=task.task_name,
            output=task.output()
        ))
        task.run()

    return task.output()



from builtins import super

from .base import AbstractTask
from Bio import SeqIO
from phigaro.context import Context

class PreprocessTask(AbstractTask):
    task_name = 'input_file'

    def __init__(self, input):
        super().__init__()
        self.input = input
        self.context = Context.instance()

    def check_fastafile(self):
        def get_users_answer(question):
            yes = ['yes', 'y']
            no = ['no', 'n']
            try:
                input_ = raw_input
            except NameError:
                input_ = input
            prompt_str = "%s %s " % (question, "[Y/n]")
            while True:
                choice = input_(prompt_str).lower()
                if (not choice) or (choice in yes):
                    return True
                if choice in no:
                    return False

        sequences_to_delete = []
        records_to_save = []
        for record in SeqIO.parse(self.input, "fasta"):
            if len(record) < 20000:
                sequences_to_delete.append(record.id)
            else:
                self.context.scaffolds_info[record.id] = len(record)
                records_to_save.append(record)
        SeqIO.write(records_to_save, self.output(), "fasta")
        del records_to_save

        if not self.config['phigaro']['delete_shorts']:
            if len(sequences_to_delete) > 0:
                print(
                    'Error! Your fasta file contains at least one sequence length < 20000.  The short sequences are: ')
                print('\n'.join(sequences_to_delete))
                if not get_users_answer('Do you want to start Phigaro without these sequences?'):
                    self.clean()
                    exit(1)

    def output(self):
        return self.file('{}.fasta'.format(self.sample))

    def run(self):
        self.check_fastafile()


from setuptools import setup

setup(name='phigaro',
      description='Phigaro is a scalable command-line tool for predictions phages and prophages '
                  'from nucleid acid sequences (including metagenomes) and '
                  'is based on phage genes HMMs and a smoothing window algorithm.',
      version="2.1.10",
      license='MIT',
      author='E.Starikova, N.Pryanichnikov, P.Tikhonova',
      author_email='hed.robin@gmail.com',
      url='https://github.com/bobeobibo/phigaro',
      packages=['phigaro',
                'phigaro.finder',
                'phigaro.cli',
                'phigaro.batch',
                'phigaro.batch.task',
                'phigaro.misc',
    'phigaro.to_html',
                ],
      package_data = {
        '': ['*.pickle', 'README.md']
      },
      entry_points={
          'console_scripts': [
              'phigaro = phigaro.cli.batch:main',
              'phigaro-setup = phigaro.cli.helper:main',
          ]
      }, install_requires=['numpy', 'six>=1.7.0', 'pandas>=0.23.4','sh', 'singleton', 'PyYAML>=5.1', 'future', 'argparse', 'numpy', 'plotly', 'bs4', 'beautifulsoup4>=4.4.0', 'lxml','biopython']
)


from builtins import input
import os
from os.path import join, exists, isfile
import sh
from future.backports.urllib.parse import urljoin
import re

from phigaro import const


class HelperException(Exception):
    message = None


class ProdigalNotFound(HelperException):
    message = "Didn't find Prodigal anywhere, " + \
              "please, download it from https://github.com/hyattpd/Prodigal/wiki/installation if you haven't done this before..."

class HMMERNotFound(HelperException):
    message = "No HMMER found. Please, download HMMER package from here: http://hmmer.org/download.html if you haven't done this before..."


def _choose_option(message, options):
    options = [
        s.rstrip()
        for s in options[:-1]
        if isfile(s.rstrip())
    ] + [options[-1]]

    if len(options) == 1:
        return options[0]

    print(message)
    while True:
        res = None
        for i, option in enumerate(options):
            print("[{}] {}".format(i + 1, option))
        option_num = input('Choose your option (Enter for {}): '.format(options[0]))
        if option_num == '':
            option_num = '1'
        if re.match(r'^\d+$', option_num):
            if int(option_num) == len(options):
                which_program = 'prodigal' if 'Prodigal' in message else 'hmmsearch'
                res = input('Please, write a path to {}: '.format(which_program))
                return res
            option_num = int(option_num) - 1
            if 0 <= option_num < len(options):
                res = options[option_num]
        if res is None:
            print('Please select valid option number ({} to {})'.format(1, len(options)))
            continue
        return res

def download_file(filename, base_url, out_dir):
    url = urljoin(base_url, filename)
    out = join(out_dir, filename)
    if exists(out):
        while True:
            overwrite = input('File {} already exists, overwrite it (Y/N)?'.format(out))
            if overwrite.upper() in {'Y', 'YES'}:
                overwrite = 'Y'
                break
            elif overwrite.upper() in {'N', 'NO'}:
                overwrite = 'N'
                break
        if overwrite == 'N':
            return

    print('Downloading {url} to {out}'.format(
        url=url,
        out=out,
    ))
    sh.wget('-O', out, url, _tty_out=True)

def download_pvogs(base_url, out_dir):
    if not exists(out_dir):
        os.makedirs(out_dir)
    # TODO: refactor 'allpvoghmms' const
    download_file('allpvoghmms', base_url, out_dir)
    download_file('allpvoghmms.h3f', base_url, out_dir)
    download_file('allpvoghmms.h3i', base_url, out_dir)
    download_file('allpvoghmms.h3m', base_url, out_dir)
    download_file('allpvoghmms.h3p', base_url, out_dir)


class SetupHelper(object):
    def __init__(self, dont_update_db):
        self.HOME = os.getenv("HOME")
        self._dont_update_db = dont_update_db

    def _locate(self, *args, **kwargs):
        if not self._dont_update_db:
            try:
                print('If you do not want to run sudo, use the following command to configure Phigaro:')
                print('       phigaro-setup --no-updatedb')
                with sh.contrib.sudo:
                    sh.updatedb()
                    self._dont_update_db = True
            except sh.ErrorReturnCode_1:
                print('Invalid password')
                exit(1)
        try:
            return [
                s.rstrip()
                for s in sh.locate(*args, **kwargs)
            ]
        except sh.ErrorReturnCode_1:
            return []

    def _find_binary(self, name, options_message, raise_if_not_found):
        locations = [sh.which(name) or ''] + self._locate('-b', '-r', '^{}$'.format(name))
        locations = [
            l.rstrip()
            for l in locations
            if l.rstrip()
        ]

        if not locations:
            # raise raise_if_not_found()
            print(raise_if_not_found)
            which_program = 'prodigal' if 'Prodigal' in raise_if_not_found else 'hmmsearch'
            locations = input('Write a full path to {}:'.format(which_program))
            return locations

        locations.append('Add another path manually')
        return _choose_option(
            message=options_message,
            options=locations,
        )

    def _setup_prodigal(self, ):
        prodigal_location = self._find_binary(name='prodigal',
                                         options_message='Please select appropriate Prodigal location',
                                         raise_if_not_found=ProdigalNotFound.message)
        return {
            'bin': prodigal_location,
        }

    def _setup_hmmer(self):
        hmmsearch_location = self._find_binary(name='hmmsearch',
                                               options_message='Please select appropriate HMMER location',
                                               raise_if_not_found=HMMERNotFound.message)
        return {
            'bin': hmmsearch_location,
        }

    def setup(self):
        prodigal_params = self._setup_prodigal()
        hmmer_params = self._setup_hmmer()

        return {
            'prodigal': {
                'bin': prodigal_params['bin'],
            },
            'hmmer': {
                'bin': hmmer_params['bin'],
                'e_value_threshold': const.DEFAULT_MAX_EVALUE,
            },
            'phigaro': {
                'window_len': const.DEFAULT_WINDOW_SIZE,
                'threshold_min': const.DEFAULT_THRESHOLD_MIN,
                'threshold_max': const.DEFAULT_THRESHOLD_MAX,
                'penalty_black': const.DEFAULT_PENALTY_BLACK,
                'penalty_white': const.DEFAULT_PENALTY_WHITE
            }
        }



from __future__ import absolute_import

import argparse
import os
from builtins import input
import sys
import yaml
from os.path import exists, dirname
import sh
from phigaro.helper import SetupHelper, HelperException, download_pvogs, download_file


def create_config(no_update_db, config_path, pvogs_dir, force):
    config_dir = dirname(config_path)
    if not exists(config_dir):
        os.makedirs(config_dir)

    if exists(config_path) and not force:
        print('Phigaro already configured')
        return

    helper = SetupHelper(no_update_db)
    try:
        config = helper.setup()
        # TODO: refactor 'allpvoghmms' const
        config['hmmer']['pvog_path'] = os.path.join(pvogs_dir, 'allpvoghmms')

        print("Found Prodigal in: {}".format(config['prodigal']['bin']))
        print("Found HMMER in: {}".format(config['hmmer']['bin']))
        print("HMMER model in: {}".format(config['hmmer']['pvog_path']))

        with open(config_path, 'w') as f:
            yaml.safe_dump(config, f, default_flow_style=False)
    except HelperException as ex:
        sys.stdout.write(ex.message+'\n')
        exit(1)
        return


def read_config_and_download_pvogs(config_path):
    if not exists(config_path):
        print("You must configure phigaro first")
        exit(1)

    with open(config_path) as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
        pvogs_dir = dirname(config['hmmer']['pvog_path'])
        print('Downloading models to {}'.format(pvogs_dir))
        download_pvogs('http://download.ripcm.com/phigaro/', pvogs_dir)

def download_test_data():
    test_data_dir = input('Please, write a full path to the directory you want test data saved to'+\
                          '(by default - working directory):')
    if test_data_dir == '':
        test_data_dir = os.getcwd()
    test_data_dir = test_data_dir.replace('\\', '/')
    test_data_dir = test_data_dir+'/test_data' if (test_data_dir[-1] != '/') else test_data_dir+'test_data'
    base_url = 'https://cdn.jsdelivr.net/gh/bobeobibo/phigaro/test_data/'
    if not exists(test_data_dir):
        os.makedirs(test_data_dir)
    download_file('Bacillus_anthracis_str_ames.fna', base_url, test_data_dir)
    download_file('Bacillus_anthracis_str_ames.phg', base_url, test_data_dir)
    download_file('Bacillus_anthracis_str_ames.phg.html', base_url, test_data_dir)

def main():
    home = os.getenv('HOME')
    phigaro_dir = os.path.join(home, '.phigaro')
    phigaro_config = os.path.join(phigaro_dir, 'config.yml')
    pvogs_dir = os.path.join(phigaro_dir, 'pvog')

    parser = argparse.ArgumentParser(prog='phigaro-setup',
                                     description="Phigaro setup helper",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-c', '--config', default=phigaro_config, help='Path to a config.yml, default is {}'.format(phigaro_config))
    parser.add_argument('-p', '--pvog', default=pvogs_dir, help='pvogs directory, default is {}'.format(pvogs_dir))
    parser.add_argument('-f','--force', action='store_true', help='Force configuration and rewrite config.yml if exists')
    parser.add_argument('--no-updatedb', action='store_true', help='Do not run sudo updatedb')
    # parser.add_argument('--download-models', action='store_true', help='Skip configuration step and download models')
    args = parser.parse_args()
    if args.force:
        sh.rm('-rf', args.config)
    create_config(
        no_update_db=args.no_updatedb,
        config_path=args.config,
        pvogs_dir=args.pvog,
        force=args.force
    )

    read_config_and_download_pvogs(args.config)
    while True:
        overwrite = input('Do you want to download test data Y/N?')
        if overwrite.upper() in {'Y', 'YES'}:
            overwrite = 'Y'
            break
        elif overwrite.upper() in {'N', 'NO'}:
            overwrite = 'N'
            break
    if overwrite == 'Y':
        download_test_data()
        print('To run phigaro on test data enter the following command:')
        print('')
        print('       phigaro -f test_data/Bacillus_anthracis_str_ames.fna -o test_data/Bacillus_anthracis_str_ames.phg -p --not-open')
        print('')

    print('The installation process is finished. If you have any questions, you can visit our github: https://github.com/bobeobibo/phigaro.')

if __name__ == '__main__':
    main()